'ToDo
'- Rx_PowerInfo -> TotalData ??

Option Explicit
'On Error Resume Next

Dim Action
Dim ReceivedData
Dim OldReceivedData
Dim MonitorCount
Dim MonitorFlag
Dim RetryCount
Dim DeviceInfoFlag
Dim PowerBufferFlag
Dim SendHandle
Dim LastMacAddress
Dim LastUpdate
Dim OldLastUpdate
Dim StickMacAddress  'Unused
Dim NetworkCode      'Unused
Dim ShortNetworkCode 'Unused
Dim Status
Dim StatusFlag
Dim MaxRetries
Dim FastTimer        'Settings Panel Timer. Set to this amount of seconds after receiving a good message
Dim SlowTimer        'Time always greater than max time for receiving any message and especially an error
Dim intTimer         'Global variable to keep track of timer to not set it when remaining the same value

Const Circle_Enabled       = 1
Const Circle_MacAddress    = 2
Const Circle_Calibration   = 3
Const Circle_LogAddress    = 4
Const Circle_DateTime      = 5
Const Circle_Usage         = 6
Const Circle_KillThreshold = 7
Const Circle_LastUpdate    = 8
Const Circle_KillFlag      = 9
Const MaxCircles           = 15
Const StatusNotOKSeconds   = 150

ReDim arrCircles(9, MaxCircles)

Call Set_Arrays()
SetPropertyValue "Plugwise Status.Status", ""
SetPropertyValue "Plugwise Status.Last Update", ""
SetPropertyValue "Plugwise Status.Action", "Initialize"
DeviceInfoFlag  = False
PowerBufferFlag = False
MonitorFlag     = "Power info"
StatusFlag      = False
FastTimer       = GetPropertyValue("Plugwise Status.Seconds") 'MilliSeconds
SlowTimer       = 10 'Seconds
MaxRetries      = 1  'Variable Counter used for retrying after an error. Normally low but high when last Circle update > StatusNotOKSeconds
OldReceivedData = GetPropertyValue("Plugwise.Received Data") 'Prevent a script trigger on Old Received Data


'-------------------------------------------------------
'- Main: Checks Received Data and Handles Send Data ----
'-------------------------------------------------------
Do
	Sleep 50
	ReceivedData = GetPropertyValue("Plugwise.Received Data")
	If OldReceivedData <> ReceivedData Then
		OldReceivedData = ReceivedData
		Call Handle_ReceivedData(ReceivedData)
	End If
	
	Action = GetPropertyValue("Plugwise Status.Action")
	If Action <> "Waiting" Then
		If Action <> "" Then
			Call Handle_SendData(Action)
			SetpropertyValue "Plugwise Status.Action", "Waiting"
			Sleep 100
		End If
	End If
	
	'Get Log Addresses for Power Buffer Info each hour on the hour
	If Minute(Now) = 2 And DeviceInfoFlag = False Then
		DeviceInfoFlag  = True
		PowerBufferFlag = False
		Call Handle_SendData("Get Device Info")
	End If
	
	'After Log Addresses are retrieved, get Power Buffer Info
	If Minute(Now) = 5 And PowerBufferFlag = False Then
		PowerBufferFlag = True
		DeviceInfoFlag  = False
		Call Handle_SendData("Get Pwr Buffer Info")
	End If
	
	If Status <> "" And StatusFlag = True Then
		Status = Time() & " - " & Status
		SetPropertyValue "Plugwise Status.Status", Status
		Status = ""
	End If
Loop


'-------------------------------------------------------
'- Handle Received Hex Data ----------------------------
'-------------------------------------------------------
Sub Handle_ReceivedData(Data)
Dim strMessage
Dim strCRC
Dim strCommand
Dim strHandle
Dim strData
Dim strFifoData
Dim nit

	Data = Replace(Data, Chr(131), "")
	Data = Split(Data, Chr(13) & Chr(10))
	
	For nit = 0 To UBound(Data)
		If Left(Data(nit), 14) = "PutFifoUnicast" Then
			strFifoData = Split(Data(nit), " : ")
			
			If UBound(strFifoData) = 2 Then
				LastMacAddress = strFifoData(2)
			Else
				LastMacAddress = "Unknown MacAddress"
			End If
		End If
		
		If Left(Data(nit), 4) = Chr(5) & Chr(5) & Chr(3) & Chr(3) Then
			If Len(Data(nit)) >= 10 Then
				strMessage = Mid(Data(nit), 5, Len(Data(nit)) - (4 + 4)) 'Strip StartBytes and Checksum
				strCRC     = Right(Data(nit), 4)
				
				If Calc_CRC(strMessage) = strCRC Then
					strCommand = Mid(strMessage, 1, 4)
					strHandle  = Mid(strMessage, 5, 4)
					strData    = Mid(strMessage, 9, Len(strMessage) - 8) '4 + 4 (CRC is already stripped)
					
					Select Case strCommand
						Case "0000": Call Rx_StickAck(strData, strHandle) 'Send 00xx
						Case "0011": Call Rx_InitAck(strData)         'Send 000A
						Case "0013": Call Rx_PowerInfo(strData)       'Send 0012 -> MonitorFlag = Power info
						Case "0024": Call Rx_DeviceInfo(strData)      'Send 0023 -> MonitorFlag = Device info
						Case "0027": Call Rx_CalibrationInfo(strData) 'Send 0026 -> MonitorFlag = Calibrate info
						Case "0049": Call Rx_PowerBufferInfo(strData) 'Send 0048 -> MonitorFlag = PwrBuffer info
					End Select
				Else
					Status = "Wrong Checksum: " & strCRC & " - Calculated CRC is: " & Calc_CRC(Right(Data(nit), 4))
				End If
			Else
				Status = "Message length too short!"
			End If
		End If
	Next
	
	'Display Circles last updated times when last update was more than 'StatusNotOKSeconds' seconds ago
	LastUpdate = ""
	
	For nit = 1 To MaxCircles
		If arrCircles(Circle_Enabled, nit) = "On" And arrCircles(Circle_Usage, nit) <> "Off" Then
			If DateDiff("s", arrCircles(Circle_LastUpdate, nit), Now) > StatusNotOKSeconds Then
				LastUpdate = LastUpdate & "Circle " & L_Pad(nit, 2, "0") & " last update is from " & arrCircles(Circle_LastUpdate, nit) & vbLF
			End If
		End If
	Next
	
	If LastUpdate = "" Then
		LastUpdate = "System Status OK"
		
		If MonitorFlag = "Power info" Then
			MaxRetries = 1
		Else
			MaxRetries = 7
		End If
	Else
		MaxRetries = 7
	End If
	
	If OldLastUpdate <> LastUpdate Then
		OldLastUpdate = LastUpdate
		SetPropertyValue "Plugwise Status.Last Update", LastUpdate
	End If
End Sub


'-------------------------------------------------------
'- Handle Send Data ------------------------------------
'-------------------------------------------------------
Sub Handle_SendData(Action)
Dim Data

	Call SetTimer("Slow Down")
	Data = Split(Action, "^")
	Select Case Data(0)
		Case "Initialize"         : RetryCount = 0 : MonitorCount = 0 : Call SendMessage("000A") : Status = "Initializing"
		Case "Get Device Info"    : RetryCount = 0 : MonitorCount = 0 : MonitorFlag = "Device info"
		Case "Get Calibrate Info" : RetryCount = 0 : MonitorCount = 0 : MonitorFlag = "Calibrate info"
		Case "Get Pwr Buffer Info": RetryCount = 0 : MonitorCount = 0 : MonitorFlag = "PwrBuffer info"
		Case "Kill Switches"      : Call KillSwitches(Data(1))
		Case "Update Circles"     : Call Set_Arrays() 'Called by Task after Enabled Checkbox changes state
		Case "Monitor"            : Call Monitor()
		Case "Switch Power"       : Call Switch_Power(CInt(Right(Data(1), 2)), Data(2)) 'Switch Power^Circle 01^Toggle
		
		Case "Toggle Status":
			If StatusFlag = False Then
				StatusFlag = True
			Else
				StatusFlag = False
				SetPropertyValue "Plugwise Status.Status", ""
			End If
			
		Case "Set Timer":
			FastTimer = GetPropertyValue("Plugwise Status.Seconds")
			
			If FastTimer > 999 Then
				SetPropertyValue "Plugwise Timer.Sleep Time", "Seconds=" & Left(FastTimer, 1) & "MilliSeconds=" & Right(FastTimer, 3)
			Else
				SetPropertyValue "Plugwise Timer.Sleep Time", "Seconds=0, MilliSeconds=" & FastTimer
			End If
	End Select
End Sub


'-------------------------------------------------------
'- Stick Acknowledge to any command followed by Info Message
'-------------------------------------------------------
Sub Rx_StickAck(Data, Handle)
Dim nit
Dim strMacAddress

	Select Case Left(Data, 4)
		Case "00C1": 'All OK
			SendHandle = Handle
			
		Case "00C2": SetTimer("Speed Up")
		Case "00D8":
			strMacAddress = Mid(Data, 5, 16)
			
			For nit = 1 To MaxCircles
				If arrCircles(Circle_MacAddress, nit) = strMacAddress Then
					arrCircles(Circle_Usage, nit)      = 0
					arrCircles(Circle_LastUpdate, nit) = Date() & " " & Time()
					arrCircles(Circle_KillFlag, nit)   = False
					SetPropertyValue "Plugwise Status.Circle " & nit & " Usage", "0.0"
					Call SetTimer("Speed Up")
					
					RetryCount = 0
					Status     = "Circle " & nit & " is switched on"
				End If
			Next
			
		Case "00DE":
			strMacAddress = Mid(Data, 5, 16)
			
			For nit = 1 To MaxCircles
				If arrCircles(Circle_MacAddress, nit) = strMacAddress Then
					arrCircles(Circle_Usage, nit)      = "Off"
					arrCircles(Circle_LastUpdate, nit) = Date() & " " & Time()
					SetPropertyValue "Plugwise Status.Circle " & nit & " Usage", "Off"
					Call SetTimer("Speed Up")
					
					RetryCount = 0
					Status     = "Circle " & nit & " is switched off"
				End If
			Next
			
		Case "00E1":
			For nit = 1 To MaxCircles
				If LastMacAddress = arrCircles(Circle_MacAddress, nit) Then
					Exit For
				End If
			Next
			
			If nit <= MaxCircles Then
				RetryCount = RetryCount + 1
				Status     = "Error: " & Data & ". Retry " & RetryCount & " on getting " & MonitorFlag & " data from Circle " & nit & " - " & GetPropertyValue("Plugwise Status.Circle " & nit & " Name")
			Else
				RetryCount = 0
				Status     = "Error: " & Data
			End If
			
			If SendHandle = Handle Then
				Call SetTimer("Speed Up")
			End If
		
		Case Else
			For nit = 1 To MaxCircles
				If LastMacAddress = arrCircles(Circle_MacAddress, nit) Then
					Exit For
				End If
			Next
			
			If nit <= MaxCircles Then
				SetPropertyValue "Plugwise Status.Circle " & nit & " Usage", "???"
				RetryCount = 0
				Status     = "Error: " & Data & " on Circle " & nit & " - " & GetPropertyValue("Plugwise Status.Circle " & nit & " Name")
			Else
				RetryCount = 0
				Status     = "Error: " & Data
			End If
			
			If SendHandle = Handle Then
				Call SetTimer("Speed Up")
			End If
	End Select
End Sub


'-------------------------------------------------------
'- Stick Acknowleldge to Initialize command ------------
'-------------------------------------------------------
Sub Rx_InitAck(Data)
Dim OnlineStatus

	StickMacAddress  = Mid(Data, 1, 16)
	OnlineStatus     = Mid(Data, 19, 2)
	NetworkCode      = Mid(Data, 21, 16)
	ShortNetworkCode = Mid(Data, 37, 4)
	
	If OnlineStatus <> "01" Then
		Status = "Network is Offline!"
	End If
End Sub


'-------------------------------------------------------
'- Circle Power Info (Current one and eight seconds) ---
'-------------------------------------------------------
Sub Rx_PowerInfo(Data)
Dim strMacAddress
Dim OneSecondData
Dim EightSecondData
Dim TotalData
Dim AverageData
Dim fltWatt
Dim intTotalWatt
Dim nit

	strMacAddress   = Left(Data, 16)
	OneSecondData   = CLng("&H" & Mid(Data, 17, 4))
	EightSecondData = CLng("&H" & Mid(Data, 21, 4))
	TotalData       = CLng("&H" & Mid(Data, 25, 8)) 'HIER NOG IETS MEE DOEN, MAAR EERST WETEN OVER WELKE TIJD DEZE DATA GELDT
	AverageData     = ((EightSecondData / 8) + OneSecondData) / 2

	For nit = 1 To MaxCircles
		If arrCircles(Circle_MacAddress, nit) = strMacAddress Then
			fltWatt = Calc_Usage(nit, AverageData, "Secs")
			
			If fltWatt <> "Error" Then
				If CLng(fltWatt) > 4500 Then
					SetPropertyValue "Plugwise Status.Circle " & nit & " Usage", "0.0"
					
					Status = "Negative Power Info received from Circle " & nit
				Else
					arrCircles(Circle_Usage, nit) = fltWatt
					SetPropertyValue "Plugwise Status.Circle " & nit & " Usage", Round(fltWatt, 1)
					
					Status = "Power Info received from Circle " & nit
				End If
				
				arrCircles(Circle_LastUpdate, nit) = Date() & " " & Time()
			Else
				Status = "Invalid calibration data for Circle " & nit
			End If
			
			Call SetTimer("Speed Up")     
			RetryCount = 0
			Exit For
		End If
	Next
	
	intTotalWatt = 0
	
	For nit = 1 To MaxCircles
		If arrCircles(Circle_Enabled, nit) = "On" Then	
			If IsNumeric(arrCircles(Circle_Usage, nit)) Then
				intTotalWatt = intTotalWatt + arrCircles(Circle_Usage, nit)
			End If
		End If
	Next
	
	SetPropertyValue "Plugwise Status.Total Usage", Round(intTotalWatt, 1)
End Sub 
        

'-------------------------------------------------------
'- Circle Device Info ----------------------------------
'-------------------------------------------------------
Sub Rx_DeviceInfo(Data)
Dim strMacAddress
Dim intYear
Dim intMonth
Dim intDays
Dim intHours
Dim intMins
Dim intMinutes
Dim strLogAddress
Dim tmpLogAddress
Dim strPower
Dim nit

	strMacAddress = Mid(Data, 1, 16)
	intYear       = 2000 + CLng("&H" & Mid(Data, 17, 2))
	intMonth      = CLng("&H" & Mid(Data, 19, 2))
	intMins       = CLng("&H" & Mid(Data, 21, 4))
	intDays       = intMins \ 60 \ 24
	intHours      = (intMins \ 60) Mod 24
	intMinutes    = intMins Mod 60
	strLogAddress = Mid(Data, 25, 8)
	strPower      = Mid(Data, 33, 2)

	For nit = 1 To MaxCircles
		If arrCircles(Circle_MacAddress, nit) = strMacAddress Then
			If strPower = "00" Then 'Switch is Off. When On, do nothing because Usage is set by Power Info
				arrCircles(Circle_Usage, nit) = "Off"
				SetPropertyValue "Plugwise Status.Circle " & nit & " Usage", "Off"
			End If
			
			arrCircles(Circle_LogAddress, nit) = strLogAddress
			arrCircles(Circle_DateTime, nit)   = intYear & "-" & intMonth & "-" & intDays & " " & intHours & ":" & intMinutes
			arrCircles(Circle_LastUpdate, nit) = Date() & " " & Time()
			Call SetTimer("Speed Up")

			RetryCount = 0			
			Status     = "Device Info received from Circle " & nit & ". " & arrCircles(Circle_DateTime, nit)
			Exit For
		End If
	Next
End Sub


'-------------------------------------------------------
'- Circle Calibration Info -----------------------------
'-------------------------------------------------------
Sub Rx_CalibrationInfo(Data)
Dim strMacAddress
Dim strGainA
Dim strGainB
Dim StrOffTot
Dim strOffRuis
Dim nit

	strMacAddress = Mid(Data, 1, 16)
	strGainA      = Mid(Data, 17, 8)
	strGainB      = Mid(Data, 25, 8)
	strOffTot     = Mid(Data, 33, 8)
	strOffRuis    = Mid(Data, 41, 8)
	
	For nit = 1 To MaxCircles
		If arrCircles(Circle_MacAddress, nit) = strMacAddress Then
			arrCircles(Circle_Calibration, nit) = strGainA & " " & strGainB & " " & strOffTot & " " & strOffRuis
			arrCircles(Circle_LastUpdate, nit)  = Date() & " " & Time()
			SetPropertyValue "Plugwise Status.Circle " & nit & " Calibration", strGainA & " " & strGainB & " " & strOffTot & " " & strOffRuis
			Call SetTimer("Speed Up")
			
			RetryCount = 0
			Status     = "Calibration Info received from Circle " & nit
			Exit For
		End If
	Next
End Sub


'-------------------------------------------------------
'- Circle Power Buffer Info ----------------------------
'-------------------------------------------------------
Sub Rx_PowerBufferInfo(Data)
Dim strMacAddress
Dim strLogHour
Dim arrHourUsage(4)
Dim nit
Dim nic
Dim nis

	nis           = 0
	strMacAddress = Mid(Data, 1, 16)

	For nit = 1 To MaxCircles
		If arrCircles(Circle_MacAddress, nit) = strMacAddress Then
			For nic = 17 To 73 Step 16
				strLogHour  = Mid(Data, nic, 8) 'OMREKENEN NAAR GOEDE TIJD?
				arrHourUsage(nis) = CInt(Calc_Usage(nit, CLng("&H" & Mid(Data, nic + 8, 8)), "Hour"))
				nis = nis + 1
			Next
			
			SetPropertyValue "Plugwise Status.Circle " & nit & " 4 Hour Usage", arrHourUsage(0) & "/" & arrHourUsage(1) & "/" & arrHourUsage(2) & "/" & arrHourUsage(3)
			arrCircles(Circle_LastUpdate, nit) = Date() & " " & Time()
			Call SetTimer("Speed Up")
			
			Status = "Power Buffer Info received from Circle " & nit
			RetryCount = 0
			Exit For
		End If
	Next
End Sub


'-------------------------------------------------------
'- Monitor Circles by requesting Power Information -----
'-------------------------------------------------------
Sub Monitor()
Dim Data
Dim LogAddress

	Data         = Split(MonitorFlag, "^")
	MonitorCount = MonitorCount + 1
	
	If RetryCount = 0 Or RetryCount >= MaxRetries Then
		RetryCount = 0
		
		If MonitorCount < MaxCircles Then
			While arrCircles(Circle_Enabled, MonitorCount) <> "On" And MonitorCount < MaxCircles
				MonitorCount = MonitorCount + 1
			Wend
		End If
		
		If MonitorCount > MaxCircles Then
			If MonitorFlag <> "Power info" Then
				MonitorFlag = "Power info" 'After one full round of Calibration or DeviceInfo revert back to monitoring
				Data(0)     = MonitorFlag
			End If
			
			MonitorCount = 1
		End If
	Else
		MonitorCount = MonitorCount - 1
	End If
		
	If arrCircles(Circle_Enabled, MonitorCount) = "On" Then     		
		Select Case Data(0)
			Case "Power info":
				If GetPropertyValue("Alarm Status.Area 1 Armed Status") <> "Disarmed" Then
					If GetPropertyValue("Plugwise Status.Circle " & MonitorCount & " Kill Switch") = "On" Then
						If arrCircles(Circle_Usage, MonitorCount) <> "Off" Then
							Call Switch_Power(MonitorCount, "Off")
						Else
							Call SetTimer("Speed Up")
						End If
					Else
						If arrCircles(Circle_Usage, MonitorCount) <> "Off" Then
							Call SendMessage("0012" & arrCircles(Circle_MacAddress, MonitorCount))
						Else
							Call SetTimer("Speed Up")
						End If
					End If
				Else
					If arrCircles(Circle_Usage, MonitorCount) <> "Off" Then
						Call SendMessage("0012" & arrCircles(Circle_MacAddress, MonitorCount))
					Else
						If arrCircles(Circle_KillFlag, MonitorCount) = True Then
							Call Switch_Power(MonitorCount, "On")
						Else
							Call SetTimer("Speed Up")
						End If
					End If
				End If
			
			Case "Device info":
				Call SendMessage("0023" & arrCircles(Circle_MacAddress, MonitorCount))
				
			Case "Calibrate info":
				Call SendMessage("0026" & arrCircles(Circle_MacAddress, MonitorCount))
			
			Case "PwrBuffer info": 
				If arrCircles(Circle_LogAddress, MonitorCount) <> "" Then
					LogAddress = ((CLng("&H" & arrCircles(Circle_LogAddress, MonitorCount)) - 278528) / 32)
					
					If LogAddress > 0 Then 
						LogAddress = LogAddress - 1 '1 LogAddress back = 4 Hours back
					End If
					
					Call SendMessage("0048" & arrCircles(Circle_MacAddress, MonitorCount) & L_Pad(Hex((LogAddress * 32) + 278528), 8, "0"))
				Else
					Call SetTimer("Speed Up")
					
					Status = "No LogAddress available for Circle " & MonitorCount
				End If
			
			Case "Kill Switches":
				While GetPropertyValue("Plugwise Status.Circle " & MonitorCount & " Kill Switch") <> "On" And MonitorCount < MaxCircles
					MonitorCount = MonitorCount + 1
				Wend
				
				If GetPropertyValue("Plugwise Status.Circle " & MonitorCount & " Kill Switch") = "On" Then
					Call Switch_Power(MonitorCount, Data(1)) 'On or Off set by Alarm script
				End If
		End Select
	Else
		Call SetTimer("Speed Up")
		RetryCount = 0 'Prevent a hang when RetryCount > 0 and current monitored Circle is Disabled
	End If
End Sub


'-------------------------------------------------------
'- Set Circle KillFlags to True when needed ------------
'-------------------------------------------------------
Sub KillSwitches(Data)
Dim nit

	RetryCount   = 0
	MonitorCount = 0
	MonitorFlag  = "Kill Switches^" & Data
	
	For nit = 1 To MaxCircles
		If arrCircles(Circle_Enabled, nit) = "On" Then
			If GetPropertyValue("Plugwise Status.Circle " & nit & " Kill Switch") = "On" Then
				If Data = "On" Then
					arrCircles(Circle_KillFlag, nit) = True
				End If
			End If
		End If
	Next
End Sub


'-------------------------------------------------------
'- Toggle Circle Power ---------------------------------
'-------------------------------------------------------
Sub Switch_Power(CircleNr, Data)

	If arrCircles(Circle_Enabled, CircleNr) = "On" Then		
		Select Case Data
			Case "On":
				Call SendMessage("0017" & arrCircles(Circle_MacAddress, CircleNr) & "01") 'Power On
				Status = "Circle " & CircleNr & " commanded On"
				
			Case "Off":
				If CInt(arrCircles(Circle_KillThreshold, CircleNr)) > 0 Then
					If IsNumeric(arrCircles(Circle_Usage, CircleNr)) Then
						If CInt(arrCircles(Circle_Usage, CircleNr)) < CInt(arrCircles(Circle_KillThreshold, CircleNr)) Then
							Call SendMessage("0017" & arrCircles(Circle_MacAddress, CircleNr) & "00") 'Power Off
							Status = "Circle " & CircleNr & " commanded Off"
						Else
							RetryCount = 0
							Status     = "Unable to switch off Circle " & CircleNr & " because device is still in use!"
							Call SetTimer("Speed Up")
						End If
					Else
						RetryCount = 0
						Status     = "Unable to switch off Circle " & CircleNr & ". Not allowed!"
						Call SetTimer("Speed Up")
					End If
				Else
					Status = "Unable to switch off Circle " & CircleNr & ". Unable to check value!"
					Call SetTimer("Speed Up")
				End If
				
			Case "Toggle":
				If GetPropertyValue("Plugwise Status.Circle " & CircleNr & " Usage") <> "Off" Then
					If IsNumeric(arrCircles(Circle_Usage, CircleNr)) Then
						If CInt(arrCircles(Circle_KillThreshold, CircleNr)) > 0 Then
							If CInt(arrCircles(Circle_Usage, CircleNr)) < CInt(arrCircles(Circle_KillThreshold, CircleNr)) Then
								Call SendMessage("0017" & arrCircles(Circle_MacAddress, CircleNr) & "00") 'Power Off
								Status = "Circle " & CircleNr & " commanded Toggle Off"
							Else
								RetryCount = 0
								Status     = "Unable to switch off Circle " & CircleNr & " because device is still in use!"
								Call SetTimer("Speed Up")
							End If
						Else
							RetryCount = 0
							Status     = "Unable to switch off Circle " & CircleNr & ". Not allowed!"
							Call SetTimer("Speed Up")
						End If
					Else
						Call SendMessage("0017" & arrCircles(Circle_MacAddress, CircleNr) & "01") 'Power On
						Status = "Unable to check value from Circle " & CircleNr & ". Therefore commanded Toggle On"
					End If
				Else
					Call SendMessage("0017" & arrCircles(Circle_MacAddress, CircleNr) & "01") 'Power On
					Status = "Circle " & CircleNr & " commanded Toggle On"
				End If		
		End Select
	Else
		Status = "Unable to switch off Circle " & CircleNr & ". Circle is disabled"
	End If
End Sub


'-------------------------------------------------------
'- Complete Message Data with CRC ----------------------
'-------------------------------------------------------
Sub SendMessage(Data)
	
	SetPropertyValue "Plugwise.Send Data", Chr(5) & Chr(5) & Chr(3) & Chr(3) & Data & Calc_CRC(Data)
End Sub


'-------------------------------------------------------
'- Set Monitor Timer Fast or according Settings Panel --
'-------------------------------------------------------
Sub SetTimer(Data)

	Select Case Data
		Case "Speed Up":
			If intTimer <> FastTimer Then
				If FastTimer > 999 Then
					SetPropertyValue "Plugwise Timer.Sleep Time", "Seconds=" & Left(FastTimer, 1) & "MilliSeconds=" & Right(FastTimer, 3)
				Else
					SetPropertyValue "Plugwise Timer.Sleep Time", "Seconds=0, MilliSeconds=" & FastTimer
				End If
				
				intTimer = FastTimer
			End If
		
		Case "Slow Down":
			If intTimer <> SlowTimer Then
				SetPropertyValue "Plugwise Timer.Sleep Time", "Seconds=" & SlowTimer
				intTimer = SlowTimer
			End If
	End Select
End Sub


'-------------------------------------------------------
'- Calc CRC 16 CCITT (X-Modem) -------------------------
'-------------------------------------------------------
Function Calc_CRC(Data)
Dim nit
Dim bit
Dim crc
	
	crc = 0
	    
	For nit = 1 To Len(Data)
		crc = crc Xor ((Asc(Mid(Data, nit, 1)) And 255) * 256)
		
		For bit = 0 To 7
			If (crc And 32768) = 32768 Then
				crc = (crc * 2) And 65535
				crc = (crc Xor &H1021)
			Else
				crc = (crc * 2) And 65535
			End If
		Next
	Next

	Calc_CRC = L_Pad(Hex(crc), 4, "0")
End Function


'-------------------------------------------------------
'- Calculate Power Usage in Watts ----------------------
'-------------------------------------------------------
Function Calc_Usage(CircleNr, Pulses, TimeSpan)
Dim CalibrationData
Dim GainA
Dim GainB
Dim OffTotal
Dim OffRuis

	Select Case TimeSpan
		Case "Secs": 'Standard
		Case "Hour" : Pulses = Pulses / 3600
	End Select
	
	CalibrationData = Split(arrCircles(Circle_Calibration, CircleNR), " ")

	If UBound(CalibrationData) = 3 Then
		GainA      = HexToFloat(CalibrationData(0))
		GainB      = HexToFloat(CalibrationData(1))
		OffTotal   = HexToFloat(CalibrationData(2))
		OffRuis    = HexToFloat(CalibrationData(3))
		Calc_Usage = 1000 * ((((((Pulses + OffRuis) ^ 2) * GainB) + ((Pulses + OffRuis) * GainA)) + OffTotal) / 468.9385193)
	Else
		Calc_Usage = "Error"
	End If	
End Function


'-------------------------------------------------------
'- Convert a 32 bit Hex string into a Float ------------
'-------------------------------------------------------
Function HexToFloat(Data)
Dim fltVal

	fltVal = (Round(((CLng("&H" & Data) And 8388607) / 2^23), 12) + 1) * (2^(((CLng("&H" & Left(Data, 6)) And 8355840) / 2^15) - 127))
	
	If (CLng("&H" & Left(Data, 2)) And 128) = 128 Then
		fltVal = fltVal * -1
	End If
	
	HexToFloat = fltVal
End Function


'-------------------------------------------------------
'- Format Data (Hour, Zones etc) to include leading 0's 
'-------------------------------------------------------
Function L_Pad(Data, Length, Char)

	L_Pad = String(Length - Len(Data), Char) & Data
End Function


'-------------------------------------------------------
'- Initialize Arrays -----------------------------------
'-------------------------------------------------------
Sub Set_Arrays()
Dim nit

	For nit = 1 To MaxCircles
		arrCircles(Circle_Enabled, nit)       = GetPropertyValue ("Plugwise Status.Circle " & nit & " Enabled")
		arrCircles(Circle_MacAddress, nit)    = GetPropertyValue ("Plugwise Status.Circle " & nit & " MacAddress")
		arrCircles(Circle_Calibration, nit)   = GetPropertyValue ("Plugwise Status.Circle " & nit & " Calibration")
		arrCircles(Circle_KillThreshold, nit) = GetPropertyValue ("Plugwise Status.Circle " & nit & " Kill Threshold")
		arrCircles(Circle_Usage, nit)         = GetPropertyValue ("Plugwise Status.Circle " & nit & " Usage")
		arrCircles(Circle_LastUpdate, nit)    = Date() & " " & Time()
		arrCircles(Circle_KillFlag, nit)      = False

		If arrCircles(Circle_Enabled, nit) = "Off" Then
			arrCircles(Circle_Usage, nit) = "---"
			SetPropertyValue "Plugwise Status.Circle " & nit & " Usage", "---"
			SetPropertyValue "Plugwise Status.Circle " & nit & " 4 Hour Usage", "0"
		Else
			If arrCircles(Circle_Usage, nit) <> "Off" Then
				arrCircles(Circle_Usage, nit) = 0
			End If
		End If
	Next
	
	Call SetTimer("Speed Up")
End Sub